import json
import struct
import warnings
from collections import OrderedDict
from io import BytesIO

def read_ndtiff_index(data, verbose=True):
    index = {}
    position = 0
    while position < len(data):
        if verbose:
            print("\rReading index... {:.1f}%       ".format(100 * (1 - (len(data) - position) / len(data))), end="")
        entry = NDTiffIndexEntry.unpack_single_index_entry(data, position)
        if entry is None:
            break
        position, axes, index_entry = entry
        index[frozenset(axes.items())] = index_entry
        if position is None:
            break
    return index


class NDTiffIndexEntry:
    EIGHT_BIT = 0
    SIXTEEN_BIT = 1
    EIGHT_BIT_RGB = 2
    TEN_BIT = 3
    TWELVE_BIT = 4
    FOURTEEN_BIT = 5
    ELEVEN_BIT = 6

    UNCOMPRESSED = 0

    def __init__(self, axes_key, pixel_type, pix_offset, image_width, image_height, md_offset, md_length, filename):
        self.axes_key = axes_key
        self.pix_offset = pix_offset
        self.image_width = image_width
        self.image_height = image_height
        self.metadata_length = md_length
        self.metadata_offset = md_offset
        self.pixel_type = pixel_type
        self.pixel_compression = self.UNCOMPRESSED
        self.metadata_compression = self.UNCOMPRESSED
        self.filename = filename
        self.data_set_finished_entry = axes_key is None

    # for backwards comapatibility when this was a dict
    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    @classmethod
    def create_finished_entry(cls):
        return cls()

    def __str__(self):
        buffer = self.as_byte_buffer()
        return buffer.getvalue().decode('iso-8859-1')

    def is_data_set_finished_entry(self):
        return self.data_set_finished_entry

    def to_essential_image_metadata(self):
        return EssentialImageMetadata(int(self.pix_width), int(self.pix_height),
                                      self.get_bit_depth(), self.is_rgb())

    @staticmethod
    def unsign_int(i):
        return i & 0xFFFFFFFF

    @staticmethod
    def read_index_map(filepath):
        with open(filepath, 'rb') as f:
            index_map = OrderedDict()
            data = f.read()
            position = 0
            while position < len(data):
                result = NDTiffIndexEntry.unpack_single_index_entry(data, position)
                if result is None:
                    break
                position, axes, index_entry = result
                index_key = NDTiffIndexEntry.serialize_axes(axes)
                index_map[index_key] = NDTiffIndexEntry(
                    index_key, index_entry.pixel_type, index_entry.pixel_offset,
                    index_entry.image_width, index_entry.image_height, index_entry.metadata_offset,
                    index_entry.metadata_length, index_entry.filename
                )
            return index_map

    @staticmethod
    def unpack_single_index_entry(data, position=0):
        """
        Unpacks a single index entry from the data buffer
        """
        (axes_length,) = struct.unpack("I", data[position: position + 4])
        if axes_length == 0:
            warnings.warn(
                "Index appears to not have been properly terminated (the dataset may still work)"
            )
            return None
        axes_str = data[position + 4: position + 4 + axes_length].decode("utf-8")
        axes = json.loads(axes_str)
        position += axes_length + 4
        (filename_length,) = struct.unpack("I", data[position: position + 4])
        filename = data[position + 4: position + 4 + filename_length].decode("utf-8")
        position += 4 + filename_length
        pixel_offset, image_width, image_height, pixel_type, pixel_compression, \
            metadata_offset, metadata_length, metadata_compression = \
            struct.unpack("IIIIIIII", data[position: position + 32])
        index_entry = NDTiffIndexEntry(axes.items, pixel_type, pixel_offset, image_width, image_height,
                                        metadata_offset, metadata_length, filename)
        position += 32
        return position, axes, index_entry


    def is_rgb(self):
        return self.pixel_type == self.EIGHT_BIT_RGB

    def get_byte_depth(self):
        return 2 if self.pixel_type in [self.SIXTEEN_BIT, self.FOURTEEN_BIT, self.TWELVE_BIT, self.TEN_BIT, self.ELEVEN_BIT] else 1

    def get_bit_depth(self):
        bit_depth_map = {
            self.EIGHT_BIT: 8,
            self.TEN_BIT: 10,
            self.TWELVE_BIT: 12,
            self.FOURTEEN_BIT: 14,
            self.SIXTEEN_BIT: 16,
            self.EIGHT_BIT_RGB: 8,
            self.ELEVEN_BIT: 11
        }
        if self.pixel_type in bit_depth_map:
            return bit_depth_map[self.pixel_type]
        else:
            raise RuntimeError("Unknown pixel type")

    def as_byte_buffer(self):
        axes = {axis_name: position for axis_name, position in self.axes_key}
        axes_key_bytes = json.dumps(axes).encode('utf-8')
        filename_bytes = self.filename.encode('utf-8')

        buffer = BytesIO()
        buffer.write(struct.pack('I', len(axes_key_bytes)))
        buffer.write(axes_key_bytes)
        buffer.write(struct.pack('I', len(filename_bytes)))
        buffer.write(filename_bytes)
        buffer.write(struct.pack('IIIIIIII', self.pix_offset, self.image_width, self.image_height,
                                 self.pixel_type, self.pixel_compression, self.metadata_offset, self.metadata_length,
                                 self.metadata_compression))
        buffer.seek(0)
        return buffer

    @staticmethod
    def deserialize_axes(s):
        return json.loads(s)

    @staticmethod
    def serialize_axes(axes):
        return json.dumps(OrderedDict(sorted(axes.items())))



"""
Abstract base class for ND image storage reading and writing

Terminiology:
- axis: a dimension of the data, e.g. 'channel', 'z', 'time'
- position: the index of a particular axis, e.g. 0, 1, 2, 3
- image_coordinates: a dictionary of axis names and their positions, e.g. {'channel': 0, 'z': 1, 'time': 2}
- image_key: a frozenset of image_coordinates, e.g. frozenset({'channel': 0, 'z': 1, 'time': 2})
- image_keys: a list of image_keys, e.g. [frozenset({'channel': 0, 'z': 1, 'time': 2}), frozenset({'channel': 0, 'z': 1, 'time': 3})]
"""
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Union, Tuple
import numpy as np
import dask
import warnings
from sortedcontainers import SortedSet
from functools import partial
import dask.array as da
import threading

from ndtiff.ndtiff_file import _POSITION_AXIS, _ROW_AXIS, _COLUMN_AXIS, _Z_AXIS, _TIME_AXIS, _CHANNEL_AXIS

# Standard order of axes when opening data. Data does not need to be acquired in this order -- this is just a convention
_AXIS_ORDER = {_ROW_AXIS: 7,
               _COLUMN_AXIS: 6,
               _POSITION_AXIS: 5,
               _TIME_AXIS: 4,
               _CHANNEL_AXIS:3,
               _Z_AXIS:2}

def _get_axis_order_key(dict_item):
    axis_name = dict_item[0]
    if axis_name in _AXIS_ORDER.keys():
        return _AXIS_ORDER[axis_name]
    else:
        return 3  # stack next to channel axes

class NDStorageAPI(ABC):
    """
    API for NDStorage classes
    """
    def __init__(self):
        # a dictionary of the types (int or str) of each axis
        # {'channel': str, 'z': int, 'time': int}
        self.axes_types = {}

        # a list of all the axes that have been seen with their values in order
        # e.g. {'time': SortedSet([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])}
        self.axes = {}

        # Metadata that applies to the entire dataset
        self.summary_metadata = None

    @abstractmethod
    def has_image(self, channel: Union[int, str] = None, z: int = None, time: int = None,
                  position: int = None, row: int = None, column: int = None,
                  **kwargs: Union[int, str]) -> bool:
        """
        Check if this image is present in the dataset.

        Parameters
        ----------
        channel : int or str, optional
            Integer index or string name of the channel (Default value = None)
        z : int, optional
            Index of z slice (Default value = None)
        time : int, optional
            Index of the time point (Default value = None)
        position : int, optional
            Index of the XY position (Default value = None)
        row : int, optional
            Index of tile row for XY tiled datasets (Default value = None)
        column : int, optional
            Index of tile column for XY tiled datasets (Default value = None)
        **kwargs :
            Names and integer positions of any other axes

        Returns
        -------
        bool :
            Indicates whether the dataset has an image matching the specifications
        """
        pass

    @abstractmethod
    def read_image(self, channel: Union[int, str] = None, z: int = None, time: int = None,
                   position: int = None, row: int = None, column: int = None,
                   **kwargs: Union[int, str]) -> Union[np.ndarray, Tuple[np.ndarray, Dict[str, Any]]]:
        """
        Read image data as numpy array.

        Parameters
        ----------
        channel : int or str, optional
            Integer index or string name of the channel (Default value = None)
        z : int, optional
            Index of z slice (Default value = None)
        time : int, optional
            Index of the time point (Default value = None)
        position : int, optional
            Index of the XY position (Default value = None)
        row : int, optional
            Index of tile row for XY tiled datasets (Default value = None)
        column : int, optional
            Index of tile column for XY tiled datasets (Default value = None)
        **kwargs :
            Names and int or str positions of any other axes

        Returns
        -------
        image : numpy array
            Image as a 2D numpy array

        """
        pass

    @abstractmethod
    def read_metadata(self, channel: Union[int, str] = None, z: int = None, time: int = None,
                      position: int = None, row: int = None, column: int = None,
                      **kwargs: Union[int, str]) -> Dict[str, Any]:
        """
        Read metadata only.

        Parameters
        ----------
        channel : int or str, optional
            Integer index or string name of the channel (Default value = None)
        z : int, optional
            Index of z slice (Default value = None)
        time : int, optional
            Index of the time point (Default value = None)
        position : int, optional
            Index of the XY position (Default value = None)
        row : int, optional
            Index of tile row for XY tiled datasets (Default value = None)
        column : int, optional
            Index of tile col for XY tiled datasets (Default value = None)
        **kwargs :
            Names and integer positions of any other axes

        Returns
        -------
        metadata : dict
            Image metadata

        """
        pass

    @abstractmethod
    def close(self):
        """
        Close the dataset, releasing any resources it holds
        """
        pass

    @abstractmethod
    def get_image_coordinates_list(self) -> List[Dict[str, Union[int, str]]]:
        """
        Return a list of the coordinates (e.g. {'channel': 'DAPI', 'z': 0, 'time': 0}) of every image in the dataset

        Returns
        -------
        list
            List of image coordinates
        """
        pass

    @abstractmethod
    def await_new_image(self, timeout=None):
        """
        Wait for a new image to arrive in the dataset

        Parameters
        ----------
        timeout : float, optional
            Maximum time to wait in seconds (Default value = None)

        Returns
        -------
        bool
            True if a new image has arrived, False if the timeout was reached
        """
        pass

    @abstractmethod
    def is_finished(self) -> bool:
        """
        Check if the dataset is finished and no more images will be added
        """
        pass



    @abstractmethod
    def as_array(self, axes: List[str] = None, stitched: bool = False,
                 **kwargs: Union[int, str]) -> 'dask.array':
        """
        Create one big Dask array with last two axes as y, x and preceding axes depending on data.
        If the dataset is saved to disk, the dask array is made up of memory-mapped numpy arrays,
        so the dataset does not need to be able to fit into RAM.
        If the data doesn't fully fill out the array (e.g. not every z-slice collected at every time point),
        zeros will be added automatically.

        To convert data into a numpy array, call np.asarray() on the returned result. However, doing so will bring the
        data into RAM, so it may be better to do this on only a slice of the array at a time.

        Parameters
        ----------
        axes : list, optional
            List of axes names over which to iterate and merge into a stacked array.
            If None, all axes will be used in PTCZYX order (Default value = None).
        stitched : bool, optional
            If True and tiles were acquired in a grid, lay out adjacent tiles next to one another
            (Default value = False)
        **kwargs :
            Names and integer positions of axes on which to slice data

        Returns
        -------
        dataset : dask array
            Dask array representing the dataset
        """
        pass

class WritableNDStorageAPI(NDStorageAPI):
    """
    API for NDStorage classes to which images can be written
    """

    @abstractmethod
    def put_image(self, coordinates, image, metadata):
        """
        Write an image to the dataset

        Parameters
        ----------
        coordinates : dict
            dictionary of axis names and positions
        image : numpy array
            image data
        metadata : dict
            image metadata

        """
        pass

    @abstractmethod
    def finish(self):
        """
        No more images will be written to the dataset
        """
        pass

    @abstractmethod
    def initialize(self, summary_metadata: dict):
        """
        Initialize the dataset with summary metadata
        """
        pass

    @abstractmethod
    def block_until_finished(self, timeout=None):
        """
        Block until the dataset is finished and all images have been written

        Parameters
        ----------
        timeout : float, optional
            Maximum time to wait in seconds (Default value = None)

        Returns
        -------
        bool
            True if the dataset is finished, False if the timeout was reached
        """
        pass


class NDStorageBase(NDStorageAPI):
    """
    Base class with helpful methods for reading and writing ND data
    """
    def __init__(self):
        super().__init__()
        self._string_axes_values = {}

        self.image_width = None
        self.image_height = None
        self.dtype = None
        self.bytes_per_pixel = None

        # for stitched datasets
        self._overlap = None
        self._full_resolution = None

        self._new_image_event = threading.Event()

    @abstractmethod
    def has_image(self, channel: Union[int, str] = None, z: int = None, time: int = None,
                  position: int = None, row: int = None, column: int = None,
                  **kwargs: Union[int, str]) -> bool:
        """
        Check if this image is present in the dataset.

        Parameters
        ----------
        channel : int or str, optional
            Integer index or string name of the channel (Default value = None)
        z : int, optional
            Index of z slice (Default value = None)
        time : int, optional
            Index of the time point (Default value = None)
        position : int, optional
            Index of the XY position (Default value = None)
        row : int, optional
            Index of tile row for XY tiled datasets (Default value = None)
        column : int, optional
            Index of tile column for XY tiled datasets (Default value = None)
        **kwargs :
            Names and integer positions of any other axes

        Returns
        -------
        bool :
            Indicates whether the dataset has an image matching the specifications
        """
        pass

    @abstractmethod
    def read_image(self, channel: Union[int, str] = None, z: int = None, time: int = None,
                   position: int = None, row: int = None, column: int = None,
                   **kwargs: Union[int, str]) -> Union[np.ndarray, Tuple[np.ndarray, Dict[str, Any]]]:
        """
        Read image data as numpy array.

        Parameters
        ----------
        channel : int or str, optional
            Integer index or string name of the channel (Default value = None)
        z : int, optional
            Index of z slice (Default value = None)
        time : int, optional
            Index of the time point (Default value = None)
        position : int, optional
            Index of the XY position (Default value = None)
        row : int, optional
            Index of tile row for XY tiled datasets (Default value = None)
        column : int, optional
            Index of tile column for XY tiled datasets (Default value = None)
        **kwargs :
            Names and int or str positions of any other axes

        Returns
        -------
        image : numpy array
            Image as a 2D numpy array

        """
        pass

    @abstractmethod
    def read_metadata(self, channel: Union[int, str] = None, z: int = None, time: int = None,
                      position: int = None, row: int = None, column: int = None,
                      **kwargs: Union[int, str]) -> Dict[str, Any]:
        """
        Read metadata only.

        Parameters
        ----------
        channel : int or str, optional
            Integer index or string name of the channel (Default value = None)
        z : int, optional
            Index of z slice (Default value = None)
        time : int, optional
            Index of the time point (Default value = None)
        position : int, optional
            Index of the XY position (Default value = None)
        row : int, optional
            Index of tile row for XY tiled datasets (Default value = None)
        column : int, optional
            Index of tile col for XY tiled datasets (Default value = None)
        **kwargs :
            Names and integer positions of any other axes

        Returns
        -------
        metadata : dict
            Image metadata

        """
        pass

    @abstractmethod
    def close(self):
        """
        Close the dataset, releasing any resources it holds
        """
        pass

    @abstractmethod
    def get_image_coordinates_list(self) -> List[Dict[str, Union[int, str]]]:
        """
        Return a list of the coordinates (e.g. {'channel': 'DAPI', 'z': 0, 'time': 0}) of every image in the dataset

        Returns
        -------
        list
            List of image coordinates
        """
        pass

    def await_new_image(self, timeout=None):
        """
        Wait for a new image to arrive in the dataset

        Parameters
        ----------
        timeout : float, optional
            Maximum time to wait in seconds (Default value = None)

        Returns
        -------
        bool
            True if a new image has arrived, False if the timeout was reached
        """
        success = self._new_image_event.wait(timeout=timeout)
        if success:
            self._new_image_event.clear()
        return success

    @abstractmethod
    def is_finished(self) -> bool:
        """
        Check if the dataset is finished and no more images will be added
        """
        pass


    ####### Implementated methods #######
    def as_array(self, axes: List[str] = None, stitched: bool = False,
                 **kwargs: Union[int, str]) -> 'dask.array':
        """
        Create one big Dask array with last two axes as y, x and preceding axes depending on data.
        If the dataset is saved to disk, the dask array is made up of memory-mapped numpy arrays,
        so the dataset does not need to be able to fit into RAM.
        If the data doesn't fully fill out the array (e.g. not every z-slice collected at every time point),
        zeros will be added automatically.

        To convert data into a numpy array, call np.asarray() on the returned result. However, doing so will bring the
        data into RAM, so it may be better to do this on only a slice of the array at a time.

        Parameters
        ----------
        axes : list, optional
            List of axes names over which to iterate and merge into a stacked array.
            If None, all axes will be used in PTCZYX order (Default value = None).
        stitched : bool, optional
            If True and tiles were acquired in a grid, lay out adjacent tiles next to one another
            (Default value = False)
        **kwargs :
            Names and integer positions of axes on which to slice data

        Returns
        -------
        dataset : dask array
            Dask array representing the dataset
        """
        if stitched:
            if self._overlap is None:
                raise Exception('This is not a stitchable dataset')
            if self._full_resolution is None:
                raise Exception('Undefinied whether this is a full resolution dataset or not')

        if None in [self.image_height, self.image_width, self.dtype, self.bytes_per_pixel]:
            raise Exception('Dataset is missing required information to create a dask array')

        if not stitched or not self._full_resolution:
            w = self.image_width
            h = self.image_height
        else:
            w = self.image_width - self._overlap[1]
            h = self.image_height - self._overlap[0]

        self._empty_tile = (
            np.zeros((h, w), self.dtype)
            if self.bytes_per_pixel != 3
            else np.zeros((h, w, 3), self.dtype)
        )

        rgb = self.bytes_per_pixel == 3 and self.dtype == np.uint8

        if axes is None:
            axes = self.axes.keys()
        axes_to_slice = kwargs
        axes_to_stack = {key: list(self.axes[key]) for key in axes if key not in kwargs.keys()}
        if stitched:
            if _ROW_AXIS in axes_to_stack:
                del axes_to_stack[_ROW_AXIS]
            if _COLUMN_AXIS in axes_to_stack:
                del axes_to_stack[_COLUMN_AXIS]
            if _ROW_AXIS in axes_to_slice:
                del axes_to_slice[_ROW_AXIS]
            if _COLUMN_AXIS in axes_to_slice:
                del axes_to_slice[_COLUMN_AXIS]

        chunks = tuple([(1,) * len(axes_to_stack[axis]) for axis in axes_to_stack.keys()])
        if stitched:
            row_values = np.array(list(self.axes[_ROW_AXIS]))
            column_values = np.array(list(self.axes[_COLUMN_AXIS]))
            chunks += (h * (np.max(row_values) - np.min(row_values) + 1),
                       w * (np.max(column_values) - np.min(column_values) + 1))
        else:
            chunks += (h, w)
        if rgb:
            chunks += (3,)

        array = da.map_blocks(
            partial(self._read_one_image_for_large_array, overlap=self._overlap, full_resolution=self._full_resolution,
                    axes_to_stack=axes_to_stack, axes_to_slice=axes_to_slice, stitched=stitched, rgb=rgb),
            dtype=self.dtype,
            chunks=chunks,
            meta=self._empty_tile
        )

        return array

    def _read_one_image_for_large_array(self, block_id, overlap, full_resolution,
                                        axes_to_stack=None, axes_to_slice=None, stitched=False, rgb=False):
        # a function that reads in one chunk of data
        axes = {key: axes_to_stack[key][block_id[i]] for i, key in enumerate(axes_to_stack.keys())}
        if stitched:
            # Combine all rows and cols into one stitched image
            # get spatial layout of position indices
            row_values = np.array(list(self.axes["row"]))
            column_values = np.array(list(self.axes["column"]))
            # fill in missing values
            row_values = np.arange(np.min(row_values), np.max(row_values) + 1)
            column_values = np.arange(np.min(column_values), np.max(column_values) + 1)
            # make nested list of rows and columns
            blocks = []
            for row in row_values:
                blocks.append([])
                for column in column_values:
                    # remove overlap between tiles
                    if not self.has_image(**axes, **axes_to_slice, row=row, column=column):
                        blocks[-1].append(self._empty_tile)
                    else:
                        tile = self.read_image(**axes, **axes_to_slice, row=row, column=column)
                        # remove half of the overlap around each tile so that that image stitches correctly
                        # only need this for full resoution because downsampled ones already have the edges removed
                        if np.any(overlap[0] > 0) and full_resolution:
                            min_index = np.floor(overlap / 2).astype(np.int_)
                            max_index = np.ceil(overlap / 2).astype(np.int_)
                            tile = tile[min_index[0]:-max_index[0], min_index[1]:-max_index[1]]
                        blocks[-1].append(tile)

            if rgb:
                image = np.concatenate([np.concatenate(row, axis=len(blocks[0][0].shape) - 2)
                        for row in blocks],  axis=0)
            else:
                image = np.array(da.block(blocks))
        else:
            if not self.has_image(**axes, **axes_to_slice):
                image = self._empty_tile
            else:
                image = self.read_image(**axes, **axes_to_slice)
        for i in range(len(axes_to_stack.keys())):
            image = image[None]
        return image

    ####### Private methods #######

    def _infer_image_properties(self, image):
        if self.dtype is None:
            # infer global dtype for as_array method
            self.image_height, self.image_width = image.shape[:2]
            self.dtype = image.dtype
            if self.dtype == np.uint8 and image.ndim == 3:
                self.bytes_per_pixel = 3 # RGB
            else:
                self.bytes_per_pixel = 2 if self.dtype == np.uint16 else 1

    def _update_axes(self, image_coordinates):
        """
        A new image has been added to the dataset, update the axes values and types
        """
        # update and ensure that all axes are either string or integer
        for axis_name, position in image_coordinates.items():
            if axis_name not in self.axes_types:
                if isinstance(position, int):
                    self.axes_types[axis_name] = int
                elif isinstance(position, str):
                    self.axes_types[axis_name] = str
                else:
                    raise RuntimeError("Axis values must be either integers or strings")
            elif type(position) != self.axes_types[axis_name]:
                raise RuntimeError("can't mix String and Integer values along an axis")

        # update the axes that have been seen
        for axis_name in image_coordinates.keys():
            if axis_name not in self.axes.keys():
                self.axes[axis_name] = SortedSet()
                self.axes_types[axis_name] = type(image_coordinates[axis_name])
            self.axes[axis_name].add(image_coordinates[axis_name])

        self._parse_string_axes_values(image_coordinates)

    def _parse_string_axes_values(self, image_coordinates):
        """
        Parse the string axes values to determine the possible values for each string axis,

        Parameters
        ----------
        image_coordinates : dict or Iterable
            A dictionary of axis names and their positions, e.g. {'channel': 0, 'z': 1, 'time': 2} or
            an iterable of frozensets of image_coordinates
        """
        # if this is a new axis_value for a string axis, add a list to populate
        for string_axis_name in [axis_name for axis_name in self.axes_types.keys() if
                                 self.axes_types[axis_name] is str]:
            if string_axis_name not in self._string_axes_values.keys():
                self._string_axes_values[string_axis_name] = []

        # if its called on just one image, make it a list of one image
        if isinstance(image_coordinates, dict):
            image_coordinates = [image_coordinates.items()]

        for single_image_coordinates in image_coordinates:
            for axis_name, axis_value in single_image_coordinates:
                if axis_name in self._string_axes_values.keys() and \
                        axis_value not in self._string_axes_values[axis_name]:
                    self._string_axes_values[axis_name].append(axis_value)


    def _parse_image_keys(self, image_keys):
        """
        Parse the image keys to determine the axes names, types, and possible values
        """
        for image_coordinates in image_keys:
            for axis_name, position in image_coordinates:
                if axis_name not in self.axes.keys():
                    self.axes[axis_name] = SortedSet()
                    self.axes_types[axis_name] = type(position)
                self.axes[axis_name].add(position)
        # Sort axes according to _AXIS_ORDER
        self.axes = dict(sorted(self.axes.items(), key=_get_axis_order_key, reverse=True))

    def _consolidate_axes(self, channel: int or str, z: int, position: int,
                          time: int, row: int, column: int, **kwargs):
        """
        Combine all the axes with standard names and custom names into a single dictionary, eliminating
        any None values. Also, convert any string-valued axes passed as ints into strings
        """
        if ('channel_name' in kwargs):
            warnings.warn('channel_name is deprecated, use "channel" instead')
            channel = kwargs['channel_name']
            del kwargs['channel_name']

        axis_positions = {'channel': channel, 'z': z, 'position': position,
                    'time': time, 'row': row, 'column': column, **kwargs}
        # ignore ones that are None
        axis_positions = {n: axis_positions[n] for n in axis_positions.keys() if axis_positions[n] is not None}
        for axis_name in axis_positions.keys():
            # convert any string-valued axes passed as ints into strings
            if self.axes_types[axis_name] == str and type(axis_positions[axis_name]) == int:
                axis_positions[axis_name] = self._string_axes_values[axis_name][axis_positions[axis_name]]

        return axis_positions



import os
import numpy as np
import threading
from ndtiff.file_io import NDTiffFileIO, BUILTIN_FILE_IO
from ndtiff.ndtiff_dataset import NDTiffDataset

class NDTiffPyramidDataset:
    """Class that opens a single NDTiffStorage multi-resolution pyramid dataset"""

    def __init__(self, dataset_path=None, file_io: NDTiffFileIO = BUILTIN_FILE_IO, summary_metadata=None):
        """
        Provides access to a NDTiffStorage pyramid dataset,
        either one currently being acquired or one on disk

        Parameters
        ----------
        dataset_path : str
            Abosolute path of top level folder of a dataset on disk
        file_io: ndtiff.file_io.NDTiffFileIO
            A container containing various methods for interacting with files.
        summary_metadata : dict
            Summary metadata, only not None for in progress datasets. Users need not call directly
        """
        self.file_io = file_io
        self._lock = threading.RLock()
        if summary_metadata is not None:
            # this dataset is a view of an active acquisition. The storage exists on the java side
            self.path = dataset_path
            self.path += "" if self.path[-1] == os.sep else os.sep
            self.summary_metadata = summary_metadata

            with self._lock:
                full_res = NDTiffDataset(dataset_path=self.path + "Full resolution" + os.sep,
                                         summary_metadata=summary_metadata, file_io=file_io)
                self.res_levels = {0: full_res}
                full_res._full_resolution = True
            # No information related higher res levels when remote storage monitoring right now

            #Copy stuff from the full res class for convenience
            self.axes = self.res_levels[0].axes

            self.overlap = (np.array([self.summary_metadata["GridPixelOverlapY"],
                                        self.summary_metadata["GridPixelOverlapX"] ]))
            self.res_levels[0].overlap = self.overlap
            # TODO maybe open other resoutions here too
            return

        # Loading from disk
        self.path = dataset_path
        self.path += "" if self.path[-1] == os.sep else os.sep
        res_dirs = [
            dI for dI in self.file_io.listdir(dataset_path) if self.file_io.isdir(self.file_io.path_join(dataset_path, dI))
        ]
        # map from downsample factor to dataset
        with self._lock:
            self.res_levels = {}
        if "Full resolution" not in res_dirs:
            #Probably won't happen because this was already checked for
            raise Exception(
                "Couldn't find full resolution directory. Is this the correct path to a dataset?"
            )

        for res_dir in res_dirs:
            res_dir_path = self.file_io.path_join(dataset_path, res_dir)
            res_level = NDTiffDataset(dataset_path=res_dir_path, file_io=self.file_io)
            if res_dir == "Full resolution":
                with self._lock:
                    self.res_levels[0] = res_level
                res_level._full_resolution = True
                # get summary metadata and index tree from full resolution image
                self.summary_metadata = res_level.summary_metadata

                self.overlap = (
                    np.array([
                            self.summary_metadata["GridPixelOverlapY"],
                            self.summary_metadata["GridPixelOverlapX"],
                        ])
                )

                self.axes = res_level.axes
                self.bytes_per_pixel = res_level.bytes_per_pixel
                self.dtype = res_level.dtype
                self.image_width = res_level.image_width
                self.image_height = res_level.image_height

            else:
                res_level._full_resolution = False
                with self._lock:
                    self.res_levels[int(np.log2(int(res_dir.split("x")[1])))] = res_level

        print("\rDataset Pyramid opened                ")


    def get_index_keys(self, res_level=0):
        """
        Return a list of every combination of axes that has a image in this dataset
        """
        return self.res_levels[res_level].get_index_keys()

    def has_new_image(self):
        """
        For datasets currently being acquired, check whether a new image has arrived since this function
        was last called, so that a viewer displaying the data can be updated.
        """
        # pass through to full resolution, since only this is monitored in current implementation
        return self.res_levels[0]

    def as_array(self, axes=None, stitched=False, res_level=None, **kwargs):
        """
        Read all data image data as one big Dask array with last two axes as y, x and preceeding axes depending on data.
        The dask array is made up of memory-mapped numpy arrays, so the dataset does not need to be able to fit into RAM.
        If the data doesn't fully fill out the array (e.g. not every z-slice collected at every time point), zeros will
        be added automatically.

        To convert data into a numpy array, call np.asarray() on the returned result. However, doing so will bring the
        data into RAM, so it may be better to do this on only a slice of the array at a time.

        Parameters
        ----------
        axes : list
            list of axes names over which to iterate and merge into a stacked array. If None, all axes will be used.
            The order of axes supplied in this list will be the order of the axes of the returned dask array
        stitched : bool
            Lay out adjacent tiles next to one another to form a larger image (Default value = False)
        res_level : int or None
            the resolution level to return. If None, return all resolutions in a list
        **kwargs :
            names and integer positions of axes on which to slice data
        Returns
        -------
        dataset : dask array
        """
        tile_shape = np.array([self.image_height, self.image_width]) - self.overlap


        def get_tile_index_from_pixel_index(pixel_index):
            """
            :param pixel_index: pixel index at relevant resolution
            """
            negative_mask = pixel_index < 0
            # positive
            tile_index = pixel_index // tile_shape
            # negative
            tile_index[negative_mask] = (-1 - (np.abs(1 + pixel_index) // tile_shape))[negative_mask]
            return tile_index


        if res_level is not None:
            return self.res_levels[res_level].as_array(axes=axes, stitched=stitched, **kwargs)
        else:
            row_values = np.array(list(self.axes["row"]))
            column_values = np.array(list(self.axes["column"]))
            pixel_extent_min = np.array([np.min(row_values), np.min(column_values)]) * tile_shape
            pixel_extent_max = np.array([np.max(row_values) + 1, np.max(column_values) + 1]) * tile_shape
            images = []
            for res_level in set(self.res_levels.keys()):
                if res_level == 0:
                    image = self.res_levels[res_level].as_array(axes=axes, stitched=stitched, **kwargs)
                    images.append(image)
                else:
                    image = self.res_levels[res_level].as_array(axes=axes, stitched=stitched, **kwargs)
                    # crop away zero padding that extends pass where data is collected
                    res_level_pixel_extent_min = (pixel_extent_min / 2 ** res_level).astype(np.int_)
                    res_level_pixel_extent_max = (pixel_extent_max / 2 ** res_level).astype(np.int_)
                    if np.min(np.stack([res_level_pixel_extent_max - res_level_pixel_extent_min,
                                        res_level_pixel_extent_max - res_level_pixel_extent_min])) < 16:
                        # Not worth it to use ones this small
                        break
                    # Subtract one to get pixel index from max extent because min extent is inclusive but max exclusive
                    min_tile_index = get_tile_index_from_pixel_index(res_level_pixel_extent_min)
                    max_tile_index = get_tile_index_from_pixel_index(res_level_pixel_extent_max - 1)
                    # get the pixel coordinates of the tiles that contain the data
                    res_level_container_extent_min = min_tile_index * tile_shape
                    res_level_container_extent_max = (max_tile_index + 1) * tile_shape
                    offset = res_level_pixel_extent_min - res_level_container_extent_min
                    extent = res_level_pixel_extent_max - res_level_pixel_extent_min
                    image = image[..., offset[0]: offset[0] + extent[0],
                            offset[1]: offset[1] + extent[1]]
                    images.append(image)


            return images


    def has_image(
        self,
        channel=None,
        z=None,
        time=None,
        position=None,
        resolution_level=0,
        row=None,
        column=None,
        **kwargs
    ):
        """Check if this image is present in the dataset

        Parameters
        ----------
        channel : int
            index of the channel, if applicable (Default value = None)
        z : int
            index of z slice, if applicable (Default value = None)
        time : int
            index of the time point, if applicable (Default value = None)
        position : int
            index of the XY position, if applicable (Default value = None)
        row : int
            index of tile row for XY tiled datasets (Default value = None)
        column : int
            index of tile column for XY tiled datasets (Default value = None)
        resolution_level :
            0 is full resolution, otherwise represents downampling of pixels
            at 2 ** (resolution_level) (Default value = 0)
        **kwargs :
            names and integer positions of any other axes

        Returns
        -------
        bool :
            indicating whether the dataset has an image matching the specifications
        """
        with self._lock:
            return self.res_levels[resolution_level].has_image(
                channel=channel,
                z=z,
                time=time,
                position=position,
                row=row,
                column=column,
                **kwargs
            )

    def read_image(
        self,
        channel=None,
        z=None,
        time=None,
        position=None,
        row=None,
        column=None,
        resolution_level=0,
        **kwargs
    ):
        """
        Read image data as numpy array

        Parameters
        ----------
        channel : int
            index of the channel, if applicable (Default value = None)
        z : int
            index of z slice, if applicable (Default value = None)
        time : int
            index of the time point, if applicable (Default value = None)
        position : int
            index of the XY position, if applicable (Default value = None)
        row : int
            index of tile row for XY tiled datasets (Default value = None)
        column : int
            index of tile col for XY tiled datasets (Default value = None)
        resolution_level :
            0 is full resolution, otherwise represents downampling of pixels
            at 2 ** (resolution_level) (Default value = 0)
        **kwargs :
            names and integer positions of any other axes

        Returns
        -------
        image : numpy array or tuple
            image as a 2D numpy array, or tuple with image and image metadata as dict

        """
        with self._lock:
            return self.res_levels[resolution_level].read_image(channel=channel,
                    z=z,
                    time=time,
                    position=position,
                    row=row,
                    column=column,
                    **kwargs)

    def read_metadata(
        self,
        channel=None,
        z=None,
        time=None,
        position=None,
        row=None,
        column=None,
        resolution_level=0,
        **kwargs
    ):
        """
        Read metadata only. Faster than using read_image to retrieve metadata

        Parameters
        ----------
        channel : int
            index of the channel, if applicable (Default value = None)
        z : int
            index of z slice, if applicable (Default value = None)
        time : int
            index of the time point, if applicable (Default value = None)
        position : int
            index of the XY position, if applicable (Default value = None)
        row : int
            index of tile row for XY tiled datasets (Default value = None)
        column : int
            index of tile col for XY tiled datasets (Default value = None)
        resolution_level :
            0 is full resolution, otherwise represents downampling of pixels
            at 2 ** (resolution_level) (Default value = 0)
        **kwargs :
            names and integer positions of any other axes

        Returns
        -------
        metadata : dict

        """
        with self._lock:
            return self.res_levels[resolution_level].read_metadata(
                    channel=channel,
                    z=z,
                    time=time,
                    position=position,
                    row=row,
                    column=column,
                    **kwargs)

    # for backwards compatibility in case of older pycromanager version, can be removed in the future
    def _add_index_entry(self, index_entry):
        self.add_index_entry(index_entry)

    def add_index_entry(self, index_entry):
        # Pass through to full res data
        return self.res_levels[0].add_index_entry(index_entry)

    def close(self):
        with self._lock:
            for res_level in self.res_levels:
                res_level.close()



name = "ndtiff"

from ndtiff.ndstorage_base import NDStorageAPI, WritableNDStorageAPI
from ndtiff._superclass import Dataset
from ndtiff.ndtiff_dataset import NDTiffDataset
from ndtiff.ndram_dataset import NDRAMDataset
from ndtiff.ndtiff_pyramid_dataset import NDTiffPyramidDataset



import numpy as np
import sys
import json
import os
import time
import struct
import warnings
from collections import OrderedDict
from io import BytesIO
from ndtiff.file_io import NDTiffFileIO, BUILTIN_FILE_IO
from ndtiff.ndtiff_index import NDTiffIndexEntry

from collections import deque
from concurrent.futures import ThreadPoolExecutor

MAJOR_VERSION = 3
MINOR_VERSION = 3

# Constants for writing files
BYTES_PER_GIG = 1073741824
MAX_FILE_SIZE = 4 * BYTES_PER_GIG

ENTRIES_PER_IFD = 13
# Required tags
WIDTH = 256
HEIGHT = 257
BITS_PER_SAMPLE = 258
COMPRESSION = 259
PHOTOMETRIC_INTERPRETATION = 262
IMAGE_DESCRIPTION = 270
STRIP_OFFSETS = 273
SAMPLES_PER_PIXEL = 277
ROWS_PER_STRIP = 278
STRIP_BYTE_COUNTS = 279
X_RESOLUTION = 282
Y_RESOLUTION = 283
RESOLUTION_UNIT = 296
MM_METADATA = 51123

SUMMARY_MD_HEADER = 2355492


_POSITION_AXIS = "position"
_ROW_AXIS = "row"
_COLUMN_AXIS = "column"
_Z_AXIS = "z"
_TIME_AXIS = "time"
_CHANNEL_AXIS = "channel"

class SingleNDTiffWriter:

    def __init__(self, directory, filename, summary_md):
        self.filename = os.path.join(directory, filename)
        self.index_map = {}
        self.next_ifd_offset_location = -1
        self.res_numerator = 1
        self.res_denominator = 1
        self.z_step_um = 1
        self.buffers = deque()
        self.first_ifd = True

        self.start_time = None

        os.makedirs(directory, exist_ok=True)
        # pre-allocate the file
        file_path = os.path.join(directory, filename)
        with open(file_path, 'wb') as f:
            f.seek(MAX_FILE_SIZE - 1)
            f.write(b'\0')
            f.flush()

        # reopen the file in binary mode
        self.file = open(file_path, 'rb+')
        # reset position to 0
        self.file.seek(0)

        self._write_mm_header_and_summary_md(summary_md)
        self.reader = SingleNDTiffReader(self.filename, summary_md=summary_md)

    def has_space_to_write(self, pixels, metadata):
        rgb = pixels.ndim == 3 and pixels.shape[2] == 3
        md_length = len(metadata)
        IFD_size = ENTRIES_PER_IFD * 12 + 4 + 16
        extra_padding = 5000000  # 5 MB extra padding
        bytes_per_pixels = self._bytes_per_image_pixels(pixels, rgb)

        size = md_length + IFD_size + bytes_per_pixels + extra_padding + self.file.tell()

        if size >= MAX_FILE_SIZE:
            return False
        return True

    def _write_mm_header_and_summary_md(self, summary_md):
        summary_md_bytes = self._get_bytes_from_string(json.dumps(summary_md))
        md_length = len(summary_md_bytes)
        header_buffer = bytearray(28)

        # 8 bytes for file header
        if sys.byteorder == 'big':
            struct.pack_into('>H', header_buffer, 0, 0x4D4D)
        else:
            struct.pack_into('<H', header_buffer, 0, 0x4949)
        struct.pack_into('<H', header_buffer, 2, 42)
        first_ifd_offset = 28 + md_length
        if first_ifd_offset % 2 == 1:
            first_ifd_offset += 1  # Start first IFD on a word
        struct.pack_into('<I', header_buffer, 4, first_ifd_offset)

        # 12 bytes for unique identifier and major version
        struct.pack_into('<III', header_buffer, 8, 483729, MAJOR_VERSION, MINOR_VERSION)

        # 8 bytes for summaryMD header and summary md length
        struct.pack_into('<II', header_buffer, 20, SUMMARY_MD_HEADER, md_length)

        for buffer in [header_buffer, summary_md_bytes]:
            self.file.write(buffer)

    def _get_bytes_from_string(self, s):
        return s.encode('utf-8')

    def finished_writing(self):
        self._write_null_offset_after_last_image()
        self.file.truncate()
        self.file.flush()
        self.file.close()

    def _write_null_offset_after_last_image(self):
        buffer = bytearray(4)
        struct.pack_into('<I', buffer, 0, 0)
        current_pos = self.file.tell()
        self.file.seek(self.next_ifd_offset_location)
        self.file.write(buffer)
        self.file.seek(current_pos)

    def write_image(self, index_key, pixels, metadata, bit_depth='auto'):
        """
        Write an image to the file

        Parameters
        ----------
        index_key : frozenset
            The key to index the image
        pixels : np.ndarray or bytearray
            The image data
        metadata : dict or str
            The metadata for the image
        bit_depth : int
            The bit depth of the image

        Returns
        -------
        NDTiffIndexEntry
            The index entry for the image
        """
        image_height, image_width = pixels.shape
        rgb = pixels.ndim == 3 and pixels.shape[2] == 3
        if bit_depth == 'auto':
            bit_depth = 8 if pixels.dtype == np.uint8 else 16
        # if metadata is a dict, serialize it to a json string and make it a utf8 byte buffer
        if isinstance(metadata, dict):
            metadata = self._get_bytes_from_string(json.dumps(metadata))
        ied = self._write_ifd(index_key, pixels, metadata, rgb, image_height, image_width, bit_depth)
        while self.buffers:
            self.file.write(self.buffers.popleft())
        self.index_map[index_key] = ied
        return ied


    def _write_ifd(self, index_key, pixels, metadata, rgb, image_height, image_width, bit_depth):
        if self.file.tell() % 2 == 1:
            self.file.seek(self.file.tell() + 1)  # Make IFD start on word

        byte_depth = 1 if isinstance(pixels, bytearray) else 2
        bytes_per_image_pixels = self._bytes_per_image_pixels(pixels, rgb)
        num_entries = 13

        # 2 bytes for number of directory entries, 12 bytes per directory entry, 4 byte offset of next IFD
        # 6 bytes for bits per sample if RGB, 16 bytes for x and y resolution, 1 byte per character of MD string
        # number of bytes for pixels
        ifd_and_bit_depth_bytes = 2 + num_entries * 12 + 4 + (6 if rgb else 0) + 16
        ifd_and_small_vals_buffer = bytearray(ifd_and_bit_depth_bytes)

        # Needed to reset to zero after last IFD
        self.next_ifd_offset_location = self.file.tell() + 2 + num_entries * 12
        bits_per_sample_offset = self.next_ifd_offset_location + 4
        x_resolution_offset = bits_per_sample_offset + (6 if rgb else 0)
        y_resolution_offset = x_resolution_offset + 8
        pixel_data_offset = y_resolution_offset + 8
        metadata_offset = pixel_data_offset + bytes_per_image_pixels

        next_ifd_offset = metadata_offset + len(metadata)
        if next_ifd_offset % 2 == 1:
            next_ifd_offset += 1  # Make IFD start on word

        buffer_position = 0
        struct.pack_into('<H', ifd_and_small_vals_buffer, buffer_position, num_entries)
        buffer_position += 2

        buffer_position += self._write_ifd_entry(ifd_and_small_vals_buffer, WIDTH, 4, 1, image_width, buffer_position)
        buffer_position += self._write_ifd_entry(ifd_and_small_vals_buffer, HEIGHT, 4, 1, image_height, buffer_position)
        buffer_position += self._write_ifd_entry(ifd_and_small_vals_buffer, BITS_PER_SAMPLE, 3, 3 if rgb else 1,
                                                 bits_per_sample_offset if rgb else byte_depth * 8, buffer_position)
        buffer_position += self._write_ifd_entry(ifd_and_small_vals_buffer, COMPRESSION, 3, 1, 1, buffer_position)
        buffer_position += self._write_ifd_entry(ifd_and_small_vals_buffer, PHOTOMETRIC_INTERPRETATION, 3, 1,
                                                 2 if rgb else 1, buffer_position)
        buffer_position += self._write_ifd_entry(ifd_and_small_vals_buffer, STRIP_OFFSETS, 4, 1, pixel_data_offset,
                                                 buffer_position)
        buffer_position += self._write_ifd_entry(ifd_and_small_vals_buffer, SAMPLES_PER_PIXEL, 3, 1, 3 if rgb else 1,
                                                 buffer_position)
        buffer_position += self._write_ifd_entry(ifd_and_small_vals_buffer, ROWS_PER_STRIP, 3, 1, image_height,
                                                 buffer_position)
        buffer_position += self._write_ifd_entry(ifd_and_small_vals_buffer, STRIP_BYTE_COUNTS, 4, 1,
                                                 bytes_per_image_pixels, buffer_position)
        buffer_position += self._write_ifd_entry(ifd_and_small_vals_buffer, X_RESOLUTION, 5, 1, x_resolution_offset,
                                                 buffer_position)
        buffer_position += self._write_ifd_entry(ifd_and_small_vals_buffer, Y_RESOLUTION, 5, 1, y_resolution_offset,
                                                 buffer_position)
        buffer_position += self._write_ifd_entry(ifd_and_small_vals_buffer, RESOLUTION_UNIT, 3, 1, 3, buffer_position)
        buffer_position += self._write_ifd_entry(ifd_and_small_vals_buffer, MM_METADATA, 2, len(metadata),
                                                 metadata_offset, buffer_position)

        struct.pack_into('<I', ifd_and_small_vals_buffer, buffer_position, next_ifd_offset)
        buffer_position += 4

        if rgb:
            struct.pack_into('<HHH', ifd_and_small_vals_buffer, buffer_position, byte_depth * 8, byte_depth * 8,
                             byte_depth * 8)
            buffer_position += 6

        struct.pack_into('<II', ifd_and_small_vals_buffer, buffer_position, self.res_numerator, self.res_denominator)
        buffer_position += 8
        struct.pack_into('<II', ifd_and_small_vals_buffer, buffer_position, self.res_numerator, self.res_denominator)
        buffer_position += 8

        self.buffers.append(ifd_and_small_vals_buffer)
        self.buffers.append(self._get_pixel_buffer(pixels, rgb))
        self.buffers.append(metadata)

        self.first_ifd = False

        # Return structured data for putting into the index entry
        pixel_type = {
            8: NDTiffIndexEntry.EIGHT_BIT,
            10: NDTiffIndexEntry.TEN_BIT,
            12: NDTiffIndexEntry.TWELVE_BIT,
            14: NDTiffIndexEntry.FOURTEEN_BIT,
            16: NDTiffIndexEntry.SIXTEEN_BIT,
            11: NDTiffIndexEntry.ELEVEN_BIT
        }.get(bit_depth, NDTiffIndexEntry.EIGHT_BIT_RGB if rgb else None)

        return NDTiffIndexEntry(index_key, pixel_type, pixel_data_offset, image_width, image_height, metadata_offset,
                                len(metadata), self.filename.split(os.sep)[-1])

    def _write_ifd_entry(self, buffer, tag, dtype, count, value, buffer_position):
        struct.pack_into('<HHII', buffer, buffer_position, tag, dtype, count, value)
        return 12

    def _get_pixel_buffer(self, pixels, rgb):
        if rgb:
            original_pix = pixels
            rgb_pix = bytearray(len(original_pix) * 3 // 4)
            num_pix = len(original_pix) // 4
            for i in range(num_pix):
                rgb_pix[i * 3] = original_pix[i * 4 + 2]
                rgb_pix[i * 3 + 1] = original_pix[i * 4 + 1]
                rgb_pix[i * 3 + 2] = original_pix[i * 4]
            return rgb_pix
        else:
            return pixels

    def _bytes_per_image_pixels(self, pixels, rgb):
        if rgb:
            return len(pixels) * 3 // 4
        else:
            if isinstance(pixels, bytearray):
                return len(pixels)
            elif isinstance(pixels, np.ndarray) and pixels.dtype == np.uint16:
                return pixels.size * 2
            else:
                raise RuntimeError("unknown pixel type")


class SingleNDTiffReader:
    """
    Class corresponsing to a single multipage tiff file
    Pass the full path of the TIFF to instantiate and call close() when finished
    """

    # file format constants
    SUMMARY_MD_HEADER = 2355492
    EIGHT_BIT_MONOCHROME = 0
    SIXTEEN_BIT_MONOCHROME = 1
    EIGHT_BIT_RGB = 2
    TEN_BIT_MONOCHROME = 3
    TWELVE_BIT_MONOCHROME = 4
    FOURTEEN_BIT_MONOCHROME = 5
    ELEVEN_BIT_MONOCHROME = 6

    UNCOMPRESSED = 0

    def __init__(self, tiff_path, file_io: NDTiffFileIO = BUILTIN_FILE_IO, summary_md=None):
        """
        tiff_path: str
            The path to a .tiff file to load
        file_io: ndtiff.file_io.NDTiffFileIO
            A container containing various methods for interacting with files.
        summary_md: dict
            If not None, this corresponds to a file that is actively being written to by an associated writer
        """
        self.file_io = file_io
        self.tiff_path = tiff_path
        self.file = self.file_io.open(tiff_path, "rb")
        if summary_md is None:
            self.summary_md, self.first_ifd_offset = self._read_header()
        else:
            self.summary_md = summary_md
            self.major_version = MAJOR_VERSION
            self.minor_version = MINOR_VERSION

    def close(self):
        """ """
        self.file.close()

    def _read_header(self):
        """
        Returns
        -------
        summary metadata : dict
        byte offsets : nested dict
            The byte offsets of TIFF Image File Directories with keys [channel_index][z_index][frame_index][position_index]
        first_image_byte_offset : int
            int byte offset of first image IFD
        """
        # read standard tiff header
        if self._read(0, 2) == b"\x4d\x4d":
            # Big endian
            if sys.byteorder != "big":
                raise Exception("Potential issue with mismatched endian-ness")
        elif self._read(0, 2) == b"\x49\x49":
            # little endian
            if sys.byteorder != "little":
                raise Exception("Potential issue with mismatched endian-ness")
        else:
            raise Exception("Endian type not specified correctly")
        if np.frombuffer(self._read(2,4), dtype=np.uint16)[0] != 42:
            raise Exception("Tiff magic 42 missing")
        first_ifd_offset = np.frombuffer(self._read(4,8), dtype=np.uint32)[0]

        # read custom stuff: header, summary md
        self.major_version = int.from_bytes(self._read(12, 16), sys.byteorder)
        self.minor_version = int.from_bytes(self._read(16, 20), sys.byteorder)

        summary_md_header, summary_md_length = np.frombuffer(self._read(20, 28), dtype=np.uint32)
        if summary_md_header != self.SUMMARY_MD_HEADER:
            raise Exception("Summary metadata header wrong")
        summary_md = json.loads(self._read(28, 28 + summary_md_length))
        return summary_md, first_ifd_offset

    def _read(self, start, end):
        """
        convert to python ints
        """
        self.file.seek(int(start), 0)
        return self.file.read(end - start)

    def read_metadata(self, index):
        return json.loads(
            self._read(
                index["metadata_offset"], index["metadata_offset"] + index["metadata_length"]
            )
        )

    def read_image(self, index_entry):
        if index_entry.pixel_type == self.EIGHT_BIT_RGB:
            bytes_per_pixel = 3
            dtype = np.uint8
        elif index_entry.pixel_type == self.EIGHT_BIT_MONOCHROME:
            bytes_per_pixel = 1
            dtype = np.uint8
        elif index_entry.pixel_type == self.SIXTEEN_BIT_MONOCHROME or \
                index_entry.pixel_type == self.TEN_BIT_MONOCHROME or \
                index_entry.pixel_type == self.TWELVE_BIT_MONOCHROME or \
                index_entry.pixel_type == self.FOURTEEN_BIT_MONOCHROME or \
                index_entry.pixel_type == self.ELEVEN_BIT_MONOCHROME:
            bytes_per_pixel = 2
            dtype = np.uint16
        else:
            raise Exception("unrecognized pixel type")
        width = index_entry.image_width
        height = index_entry.image_height
        data = self._read(index_entry.pix_offset, index_entry.pix_offset + width * height * bytes_per_pixel)
        pixels = np.frombuffer(data, dtype=dtype)
        image = pixels.reshape([height, width, 3] if bytes_per_pixel == 3 else [height, width])
        return image



import os
import threading
import time

import numpy as np
import warnings
import re

from ndtiff.file_io import NDTiffFileIO, BUILTIN_FILE_IO
from ndtiff.ndtiff_file import SingleNDTiffReader
from ndtiff.ndtiff_file import _CHANNEL_AXIS
from ndtiff.ndtiff_index import NDTiffIndexEntry, read_ndtiff_index

from ndtiff.ndtiff_file import SingleNDTiffWriter, MAJOR_VERSION, MINOR_VERSION

from ndtiff.ndstorage_base import WritableNDStorageAPI, NDStorageBase

class NDTiffDataset(NDStorageBase, WritableNDStorageAPI):
    """
    Class that opens a single NDTiff dataset
    """

    def __init__(self, dataset_path=None, file_io: NDTiffFileIO = BUILTIN_FILE_IO, summary_metadata=None,
                 name=None, writable=False, **kwargs):
        """
        Provides access to an NDTiffStorage dataset,
        either one currently being acquired or one on disk

        Parameters
        ----------
        dataset_path : str
            Abosolute path of top level folder of a dataset on disk
        file_io: ndtiff.file_io.NDTiffFileIO
            A container containing various methods for interacting with files.
        summary_metadata : dict
            Summary metadata for a dataset that is currently being written by another process
        name : str
            Name of the dataset if writing a new dataset
        writable : bool
            Whether it is a new dataset being written to disk
        """
        super().__init__()

        self.file_io = file_io
        self._lock = threading.RLock()
        if writable:
            self.major_version = MAJOR_VERSION
            self.minor_version = MINOR_VERSION
            self._index_file = None
        if summary_metadata is not None or writable:
            # this dataset is either:
            #   - a view of an active acquisition. Image data is being written by code on the Java side
            #   - a new dataset being written to disk
            self.index = {}
            self._readers_by_filename = {}
            self._summary_metadata = summary_metadata
            self._writable = writable
            self.current_writer = None
            self.file_index = 0
            self.name = name
            self._write_pending_images = {}
            self._finished_event = threading.Event()

            if writable and name is not None:
                # create a folder to hold the new Tiff files
                self.path = _create_unique_acq_dir(dataset_path, name)
            else:
                self.path = dataset_path
                self.path += "" if self.path[-1] == os.sep else os.sep
        else:
            self._write_pending_images = None
            self._writable = False
            self.path = dataset_path
            self.path += "" if self.path[-1] == os.sep else os.sep
            print("\rReading index...          ", end="")
            with self.file_io.open(os.sep.join((self.path, "NDTiff.index")), "rb") as index_file:
                data = index_file.read()
                self.index = read_ndtiff_index(data)

            tiff_names = [
                self.file_io.path_join(self.path, tiff) for tiff in self.file_io.listdir(self.path) if tiff.endswith(".tif")
            ]
            self._readers_by_filename = {}
            self.summary_metadata = {}
            self.major_version, self.minor_version = (0, 0)
            # Count how many files need to be opened
            num_tiffs = 0
            count = 0
            for file in self.file_io.listdir(self.path):
                if file.endswith(".tif"):
                    num_tiffs += 1
            # populate list of readers and tree mapping indices to readers
            for tiff in tiff_names:
                print("\rOpening file {} of {}...".format(count + 1, num_tiffs), end="")
                count += 1
                new_reader = SingleNDTiffReader(tiff, file_io=self.file_io)
                self._readers_by_filename[tiff.split(os.sep)[-1]] = new_reader
                # Should be the same on every file so resetting them is fine
                self.major_version, self.minor_version = new_reader.major_version, new_reader.minor_version

            if len(self._readers_by_filename) > 0:
                self.summary_metadata = list(self._readers_by_filename.values())[0].summary_md

            # TODO: make overlap non-public in a future version
            self.overlap = (
                np.array([self.summary_metadata["GridPixelOverlapY"], self.summary_metadata["GridPixelOverlapX"]])
                if "GridPixelOverlapY" in self.summary_metadata else None)
            self._overlap = self.overlap

            self._parse_image_keys(self.index.keys())

            # figure out the mapping of channel name to position by reading image metadata
            self._channels = {}
            self._update_channel_names(self.index.keys())

            # get information about image width and height, assuming that they are consistent for whole dataset
            # (which is not necessarily true but convenient when it is)
            if len(self.index) > 0:
                with self._lock:
                    first_index = list(self.index.values())[0]
                self._parse_essential_image_metadata(first_index)

            print("\rDataset opened                ")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def get_channel_names(self):
        """
        :return: list of channel names (strings)
        """
        return list(self._channels.keys())

    def has_image(self, channel=None, z=None, time=None, position=None, row=None, column=None, **kwargs):
        with self._lock:
            axes = self._consolidate_axes(channel, z, position, time, row, column, **kwargs)

            if self._write_pending_images is not None and len(self._write_pending_images) > 0:
                if frozenset(axes.items()) in self._write_pending_images:
                    return True

            return self._does_have_image(axes)

    def read_image(self, channel=None, z=None, time=None, position=None, row=None, column=None, **kwargs):
        with self._lock:
            axes = self._consolidate_axes(channel, z, position, time, row, column, **kwargs)

            if self._write_pending_images is not None and len(self._write_pending_images) > 0:
                if frozenset(axes.items()) in self._write_pending_images:
                    return self._write_pending_images[frozenset(axes.items())][0]

            return self._do_read_image(axes)

    def read_metadata(self, channel=None, z=None, time=None, position=None, row=None, column=None, **kwargs):
        with self._lock:
            axes = self._consolidate_axes(channel, z, position, time, row, column, **kwargs)

            if self._write_pending_images is not None and len(self._write_pending_images) > 0:
                if frozenset(axes.items()) in self._write_pending_images:
                    return self._write_pending_images[frozenset(axes.items())][1]

            return self._do_read_metadata(axes)

    def put_image(self, coordinates, image, metadata):
        if not self._writable:
            raise RuntimeError("Cannot write to a read-only dataset")

        # add to write pending images
        self._write_pending_images[frozenset(coordinates.items())] = (image, metadata)

        self._update_axes(coordinates)
        self._update_channel_names(coordinates)
        self._infer_image_properties(image)

        # Update viewer as soon as image is ready in RAM
        self._new_image_event.set()

        # Create a new file if needed
        if self.current_writer is None:
            filename = 'NDTiffStack.tif'
            if self.name is not None:
                filename = self.name + '_' + filename
            self.current_writer = SingleNDTiffWriter(self.path, filename, self._summary_metadata)
            self.file_index += 1
            # create the index file
            self._index_file = open(os.path.join(self.path, "NDTiff.index"), "wb")
        elif not self.current_writer.has_space_to_write(image, metadata):
            self.current_writer.finished_writing()
            filename = 'NDTiffStack_{}.tif'.format(self.file_index)
            if self.name is not None:
                filename = self.name + '_' + filename
            self.current_writer = SingleNDTiffWriter(self.path, filename, self._summary_metadata)
            self.file_index += 1

        index_data_entry = self.current_writer.write_image(frozenset(coordinates.items()), image, metadata)
        # create readers and update axes
        self.add_index_entry(index_data_entry, new_image_updates=False)
        # write the index to disk
        self._index_file.write(index_data_entry.as_byte_buffer().getvalue())
        # remove from pending images
        del self._write_pending_images[frozenset(coordinates.items())]

    def finish(self):
        if self.current_writer is not None:
            self.current_writer.finished_writing()
            self.current_writer = None
        if self._index_file is not None: # if no images were written, it never got opened
            self._index_file.close()
        self._finished_event.set()

    def is_finished(self) -> bool:
        return self._finished_event.is_set()

    def initialize(self, summary_metadata: dict):
        # called for new storage
        self._summary_metadata = summary_metadata

    def block_until_finished(self, timeout=None):
        return self._finished_event.wait(timeout=timeout)

    def get_image_coordinates_list(self):
        frozen_set_list = list(self.index.keys())
        # convert to dict
        return [{axis_name: position for axis_name, position in key} for key in frozen_set_list]

    # TODO: remove this in a future version
    def get_index_keys(self):
        """
        Return a list of every combination of axes that has a image in this dataset
        """
        warnings.warn("get_index_keys is deprecated, use get_image_coordinates_list instead", DeprecationWarning)
        return self.get_image_coordinates_list()

    def add_index_entry(self, data, new_image_updates=True):
        """
        Add entry for an image that has been received and is now on disk
        This is used when the data is being written outside this class (i.e. by Java code)

        Parameters
        ----------
        data : bytes or NDTiffIndexEntry
            binary data for the index entry
        new_image_updates : bool
            whether to signal that a new image is ready
        """
        with self._lock:
            if isinstance(data, NDTiffIndexEntry):
                index_entry = data
                # reconvert to dict from frozenset
                image_coordinates = {axis_name: position for axis_name, position in index_entry.axes_key}
                self.index[index_entry.axes_key] = index_entry
            else:
                _, image_coordinates, index_entry = NDTiffIndexEntry.unpack_single_index_entry(data)
                self.index[frozenset(image_coordinates.items())] = index_entry

            if index_entry.filename not in self._readers_by_filename:
                new_reader = SingleNDTiffReader(os.path.join(self.path, index_entry.filename), file_io=self.file_io)
                self._readers_by_filename[index_entry.filename] = new_reader
                # Should be the same on every file so resetting them is fine
                self.major_version, self.minor_version = new_reader.major_version, new_reader.minor_version

            self._parse_essential_image_metadata(index_entry)

            if new_image_updates:
                self._update_axes(image_coordinates)
                self._update_channel_names(image_coordinates)
                self._new_image_event.set()



        return image_coordinates

    def _update_channel_names(self, image_coordinates):
        """
        As of NDTiff 3.2, axes are allowed to take string values: e.g. {'channel': 'DAPI'}
        This is allowed on any axis. This function returns a tuple of possible values along
        the string axis in order to be able to interconvert integer values and string values.

        param single_image_coordinates: if provided, only parse this newly added entry
        """
        # iterate through the key_combos for each image
        if self.major_version >= 3 and self.minor_version >= 2:
            if _CHANNEL_AXIS in self._string_axes_values:
                self._channels = {name: self._string_axes_values[_CHANNEL_AXIS].index(name)
                                  for name in self._string_axes_values[_CHANNEL_AXIS]}
        else:
            # before string-valued axes were allowed in NDTiff 3.1
            if 'ChNames' in self.summary_metadata:
                # It was created by a MM MDA/Clojure acquistiion engine
                self._channels = {name: i for i, name in enumerate(self.summary_metadata['ChNames'])}
            else:
                # AcqEngJ
                if _CHANNEL_AXIS in self.axes.keys():
                    for key in self.index.keys():
                        single_image_coordinates = {axis: position for axis, position in key}
                        if (_CHANNEL_AXIS in single_image_coordinates.keys()
                                    and single_image_coordinates[_CHANNEL_AXIS] not in self._channels.values()):
                            channel_name = self.read_metadata(**single_image_coordinates)["Channel"]
                            self._channels[channel_name] = single_image_coordinates[_CHANNEL_AXIS]
                        if len(self._channels.values()) == len(self.axes[_CHANNEL_AXIS]):
                            break

    def _parse_essential_image_metadata(self, index_entry):
        """
        Read through first index to get some global metadata about images
        This is used for the as_array method to determine the dtype and shape of the array
        It assumes that the image width and height and dtype are consistent across the dataset,
        which is not necessarily true, but as_array doesn't work if its not anyway
        """
        if self.dtype is None:
            if index_entry["pixel_type"] == SingleNDTiffReader.EIGHT_BIT_RGB:
                self.bytes_per_pixel = 3
                self.dtype = np.uint8
            elif index_entry["pixel_type"] == SingleNDTiffReader.EIGHT_BIT_MONOCHROME:
                self.bytes_per_pixel = 1
                self.dtype = np.uint8
            elif index_entry["pixel_type"] == SingleNDTiffReader.SIXTEEN_BIT_MONOCHROME or \
                    index_entry["pixel_type"] == SingleNDTiffReader.FOURTEEN_BIT_MONOCHROME or \
                    index_entry["pixel_type"] == SingleNDTiffReader.TWELVE_BIT_MONOCHROME or \
                    index_entry["pixel_type"] == SingleNDTiffReader.TEN_BIT_MONOCHROME or \
                    index_entry["pixel_type"] == SingleNDTiffReader.ELEVEN_BIT_MONOCHROME:
                self.bytes_per_pixel = 2
                self.dtype = np.uint16
            else:
                warnings.warn("Unknown pixel type: {}\nAssuming 16 bit".format(index_entry["pixel_type"]))
                self.bytes_per_pixel = 2
                self.dtype = np.uint16

            self.image_width = index_entry["image_width"]
            self.image_height = index_entry["image_height"]

    def _does_have_image(self, axes):
        key = frozenset(axes.items())
        return key in self.index

    def _do_read_image(self, axes,):
        # determine which reader contains the image
        key = frozenset(axes.items())
        if key not in self.index:
            raise Exception("image with keys {} not present in data set".format(key))
        index = self.index[key]
        reader = self._readers_by_filename[index["filename"]]
        return reader.read_image(index)

    def _do_read_metadata(self, axes):
        """

        Parameters
        ----------
        axes : dict

        Returns
        -------
        image_metadata
        """
        key = frozenset(axes.items())
        if key not in self.index:
            raise Exception("image with keys {} not present in data set".format(key))
        index = self.index[key]
        reader = self._readers_by_filename[index["filename"]]
        return reader.read_metadata(index)

    def close(self):
        for reader in self._readers_by_filename.values():
            reader.close()



def _create_unique_acq_dir(root, prefix):
    if not os.path.exists(root):
        os.makedirs(root)

    pattern = re.compile(re.escape(prefix) + r"_(\d+).*", re.IGNORECASE)
    max_number = 0

    for acq_dir in os.listdir(root):
        match = pattern.match(acq_dir)
        if match:
            try:
                number = int(match.group(1))
                max_number = max(max_number, number)
            except ValueError:
                pass

    new_dir_name = f"{prefix}_{max_number + 1}"
    new_dir_path = os.path.join(root, new_dir_name)
    os.makedirs(new_dir_path)

    return new_dir_path


import numpy as np
from ndtiff import Dataset
import os


def test_v2_data(test_data_path):
    data_path = os.path.join(test_data_path, "v2", "ndtiffv2.0_test")
    dataset = Dataset(data_path)
    assert(np.mean(dataset.as_array()) > 0)

def test_v2_data_axis_sorting(test_data_path):
    data_path = os.path.join(test_data_path, "v2", "ndtiffv2.0_test")
    dataset = Dataset(data_path)
    num_timepoints = len(dataset.axes['time'])
    num_channels = len(dataset.axes['channel'])
    data = dataset.as_array(axes=None)
    assert data.shape[:-2] == (num_timepoints, num_channels)

def test_v3_data(test_data_path):
    data_path = os.path.join(test_data_path, 'v3', 'ndtiffv3.0_test')
    dataset = Dataset(data_path)
    assert(np.mean(dataset.as_array()) > 0)
    for channel_name, correct_channel_name in zip(dataset.get_channel_names(), ['DAPI', 'FITC']):
        assert(channel_name == correct_channel_name)

def test_v2_stitched_data(test_data_path):
    data_path = os.path.join(test_data_path, "v2", "ndtiffv2.0_stitched_test")
    dataset = Dataset(data_path)
    stitched = np.array(dataset.as_array(stitched=True))
    assert(stitched[..., 0, 0] > 0)
    assert(stitched[..., -1, -1] > 0)
    assert(stitched[..., 0, -1] == 0)
    assert(stitched[..., -1, 0] == 0)

def test_v3_stitched_data(test_data_path):
    data_path = os.path.join(test_data_path, 'v3', 'ndtiffv3.0_stitched_test')
    dataset = Dataset(data_path)
    stitched = np.array(dataset.as_array(stitched=True, res_level=0))
    assert(stitched[..., 0, 0] > 0)
    assert(stitched[..., -1, -1] > 0)
    assert(stitched[..., 0, -1] == 0)
    assert(stitched[..., -1, 0] == 0)

def test_v3_magellan_explore(test_data_path):
    data_path = os.path.join(test_data_path, 'v3', 'Magellan_expolore_multi_channel')
    dataset = Dataset(data_path)
    stitched = np.array(dataset.as_array(stitched=True, res_level=0))

def test_v3_magellan_explore_negative_and_overwritten(test_data_path):
    data_path = os.path.join(test_data_path, 'v3', 'Magellan_expolore_negative_indices_and_overwritten')
    dataset = Dataset(data_path)
    stitched = np.array(dataset.as_array(stitched=True, res_level=0))

def test_v3_non_ctzp_axes(test_data_path):
    data_path = os.path.join(test_data_path, 'v3', 'Nonstandard_axis_names')
    dataset = Dataset(data_path)
    stitched = np.array(dataset.as_array())

def test_v3_mm_mda(test_data_path):
    data_path = os.path.join(test_data_path, 'v3', 'mm_mda_tcz_15')
    dataset = Dataset(data_path)
    stitched = np.array(dataset.as_array())

def test_v3_2_multichannel(test_data_path):
    data_path = os.path.join(test_data_path, 'v3', 'ndtiff3.2_multichannel')
    dataset = Dataset(data_path)
    assert(np.mean(dataset.read_image(channel='DAPI', time=0, z=1)) > 0)
    for channel_name, correct_channel_name in zip(dataset.get_channel_names(), ['DAPI', 'FITC']):
        assert(channel_name == correct_channel_name)

def test_v3_2_multichannel_axis_sorting(test_data_path):
    data_path = os.path.join(test_data_path, 'v3', 'ndtiff3.2_multichannel')
    dataset = Dataset(data_path)
    num_timepoints = len(dataset.axes['time'])
    num_channels = len(dataset.axes['channel'])
    num_slices = len(dataset.axes['z'])
    data = dataset.as_array(axes=None)
    assert data.shape[:-2] == (num_timepoints, num_channels, num_slices)

def test_v3_2_monochrome(test_data_path):
    data_path = os.path.join(test_data_path, 'v3', 'ndtiff3.2_monochrome')
    dataset = Dataset(data_path)
    assert (np.mean(dataset.read_image(time=0, z=1)) > 0)

def test_v3_2_magellan_rgb_explore(test_data_path):
    data_path = os.path.join(test_data_path, 'v3', 'ndtiff3.2_magellan_explore_rgb')
    dataset = Dataset(data_path)
    assert(np.sum(np.array(dataset.as_array(stitched=True)[0])) > 0)

def test_v3_2_11bit_data(test_data_path):
    data_path = os.path.join(test_data_path, 'v3', 'ndtiff3.2_11bit_1')
    dataset = Dataset(data_path)
    assert(dataset.read_metadata(time=0)['BitDepth'] == 11)
    assert(dataset.read_image(time=0).dtype == np.uint16)

def test_v3_2_12bit_pycromanager_data(test_data_path):
    data_path = os.path.join(test_data_path, 'v3', '12_bit_pycromanager_mda')
    dataset = Dataset(data_path)
    assert(dataset.read_image(time=0, channel='DAPI').dtype == np.uint16)

def test_v3_2_no_magellan_explore_channels(test_data_path):
    data_path = os.path.join(test_data_path, 'v3', 'no_magellan_explore_multi_channel')
    dataset = Dataset(data_path)
    assert(np.sum(np.array(dataset.as_array(stitched=True)[0])) > 0)

def test_v3_2_no_magellan_explore_no_channels(test_data_path):
    data_path = os.path.join(test_data_path, 'v3', 'no_magellan_explore_no_channel')
    dataset = Dataset(data_path)
    assert(np.sum(np.array(dataset.as_array(stitched=True)[0])) > 0)

def test_labeled_positions(test_data_path):
    """
    This dataset was acquired with the following events:

    events = [
        {'axes': {'position': 'Pos2'}, 'x': 0, 'y': 0},
        {'axes': {'position': 'Pos0'}, 'x': 0, 'y': 1},
        {'axes': {'position': 'Pos1'}, 'x': 0, 'y': 2}
    ]

    Further, images at position ('Pos2', 'Pos0', 'Pos1') have (0, 0) pixels
    equal to (0, 1, 2).

    Here we test that the axis and image order of the dataset is as expected.
    """

    data_path = os.path.join(test_data_path, 'v3', 'labeled_positions_1')
    position_names = ('Pos2', 'Pos0', 'Pos1')

    dataset = Dataset(data_path)
    data = np.asarray(dataset.as_array())

    # ndtiff sorts the dataset axes. Check that the position axis is correctly
    # sorted
    assert dataset.axes['position'] == set(('Pos0', 'Pos1', 'Pos2'))

    # Check that the image metadata includes the position name
    for position_name in position_names:
        metadata = dataset.read_metadata(position=position_name)
        assert metadata['PositionName'] == position_name

    # Check that data is in the ('Pos0', 'Pos1', 'Pos2') order
    pixel_00 = (1, 2, 0)
    for idx, image in enumerate(data):
        assert image[0, 0] == pixel_00[idx]

def test_unordered_z_axis(test_data_path):
    """
    This dataset was acquired with the following events:

    events1 = [{'axes': {'z': z_idx}, 'z': z_idx} for z_idx in range(10)]
    events2 = [{'axes': {'z': z_idx}, 'z': z_idx} for z_idx in range(-10, 0)]

    and the following image process hook function:

    def fun(image, metadata):
        if not hasattr(fun, "idx"):
            fun.idx = 0

        image[0, 0] = np.uint16(fun.idx)

        fun.idx += 1

        return image, metadata

    Here we test that the axis and image order of the dataset is as expected.
    """

    data_path = os.path.join(test_data_path, 'v3', 'unordered_z_1')
    acquisition_z_axis = list(range(10)) + list(range(-10, 0))
    sorting_index = np.argsort(acquisition_z_axis)
    sorted_z_axis = np.sort(acquisition_z_axis)

    dataset = Dataset(data_path)
    data = np.asarray(dataset.as_array())

    # ndtiff sorts the dataset axes. Check that the position axis is correctly
    # sorted
    assert dataset.axes['z'] == set(sorted_z_axis)

    # Check that data is in the sorted(acquisition_z_axis) order
    pixel_00 = sorting_index
    for idx, image in enumerate(data):
        assert image[0, 0] == pixel_00[idx]



from typing import List, Dict, Union

import threading
from ndtiff.ndstorage_base import WritableNDStorageAPI, NDStorageBase


class NDRAMDataset(NDStorageBase, WritableNDStorageAPI):
    """
    A class for holding data in RAM
    Implements the methods needed to be a DataSink for AcqEngPy
    """

    def __init__(self):
        super().__init__()
        self.images = {}
        self.image_metadata = {}
        self._finished_event = threading.Event()

    def initialize(self, summary_metadata: dict):
        self.summary_metadata = summary_metadata

    def block_until_finished(self, timeout=None):
        return self._finished_event.wait(timeout=timeout)

    def finish(self):
        self._finished_event.set()

    def is_finished(self) -> bool:
        return self._finished_event.is_set()

    def put_image(self, coordinates, image, metadata):
        key = frozenset(coordinates.items())
        self.images[key] = image
        self.image_metadata[key] = metadata
        self._update_axes(coordinates)
        self._new_image_event.set()

    def get_image_coordinates_list(self) -> List[Dict[str, Union[int, str]]]:
        frozen_set_list = list(self.images.keys())
        # convert to dict
        return [{axis_name: position for axis_name, position in key} for key in frozen_set_list]

    #### ND Storage API ####
    def close(self):
        self.images = {}
        self.image_metadata = {}
        self.axes = {}

    def has_image(self, channel=None, z=None, time=None, position=None, row=None, column=None, **kwargs):
        axes = self._consolidate_axes(channel, z, position, time, row, column, **kwargs)
        key = frozenset(axes.items())
        return key in self.images.keys()

    def read_image(self, channel=None, z=None, time=None, position=None, row=None, column=None, **kwargs):
        axes = self._consolidate_axes(channel, z, position, time, row, column, **kwargs)
        key = frozenset(axes.items())
        if key not in self.images.keys():
            raise Exception("image with keys {} not present in data set".format(key))
        return self.images[key]

    def read_metadata(self, channel=None, z=None, time=None, position=None, row=None, column=None, **kwargs):
        axes = self._consolidate_axes(channel, z, position, time, row, column, **kwargs)
        key = frozenset(axes.items())
        if key not in self.images.keys():
            raise Exception("image with keys {} not present in data set".format(key))
        return self.image_metadata[key]




import numpy as np
import os
import shutil
from ndtiff.ndtiff_file import SingleNDTiffWriter, SingleNDTiffReader
from ndtiff.ndtiff_dataset import NDTiffDataset
from ndtiff.ndram_dataset import NDRAMDataset
import pytest

@pytest.fixture(scope="function")
def test_data_path(tmp_path_factory):
    data_path = tmp_path_factory.mktemp('writer_tests')
    for f in os.listdir(data_path):
        os.remove(os.path.join(data_path, f))
    yield str(data_path)
    shutil.rmtree(data_path)

def test_write_single_file(test_data_path):
    """
    Create a single NDTiff file and read it back in
    """
    filename = 'test_write_single_file.tif'
    writer = SingleNDTiffWriter(test_data_path, filename, summary_md={})

    image_height = 256
    image_width = 256
    pixels = np.arange(image_height * image_width, dtype=np.uint16).reshape((image_height, image_width))

    index_key = frozenset({'time': 0}.items())
    index_entry = writer.write_image(index_key, pixels, {})
    writer.finished_writing()

    # read the file back in
    single_reader = SingleNDTiffReader(os.path.join(test_data_path, filename))
    read_pixels = single_reader.read_image(index_entry)
    assert np.all(read_pixels == pixels)


def test_write_full_dataset(test_data_path):
    """
    Write an NDTiff dataset and read it back in, testing pixels and metadata
    """
    full_path = os.path.join(test_data_path, 'test_write_full_dataset')
    dataset = NDTiffDataset(full_path, summary_metadata={}, writable=True)

    image_height = 256
    image_width = 256
    images = []
    for time in range(10):
        pixels = np.ones(image_height * image_width, dtype=np.uint16).reshape((image_height, image_width)) * time
        images.append(pixels)
    for time in range(10):
        axes = {'time': time}
        dataset.put_image(axes, images[time], {'time_metadata': time})

    dataset.finish()

    # read the file back in
    dataset = NDTiffDataset(full_path)
    for time in range(10):
        pixels = np.ones(image_height * image_width, dtype=np.uint16).reshape((image_height, image_width)) * time
        axes = {'time': time}
        read_image = dataset.read_image(**axes)
        assert np.all(read_image == pixels)
        assert dataset.read_metadata(**axes) == {'time_metadata': time}

def test_write_full_dataset_RAM():
    dataset = NDRAMDataset()

    image_height = 256
    image_width = 256
    images = []
    for time in range(10):
        pixels = np.ones(image_height * image_width, dtype=np.uint16).reshape((image_height, image_width)) * time
        images.append(pixels)
    for time in range(10):
        axes = {'time': time}
        dataset.put_image(axes, images[time], {'time_metadata': time})

    dataset.finish()

    for time in range(10):
        pixels = np.ones(image_height * image_width, dtype=np.uint16).reshape((image_height, image_width)) * time
        axes = {'time': time}
        read_image = dataset.read_image(**axes)
        assert np.all(read_image == pixels)
        assert dataset.read_metadata(**axes) == {'time_metadata': time}






# import numpy as np
# import os
# import shutil
# from ndtiff.ndtiff_file import SingleNDTiffWriter, SingleNDTiffReader
# from ndtiff.ndtiff_dataset import NDTiffDataset
# import pytest
# import time
# from tqdm import tqdm
#
# def writing_speed_test(test_data_path):
#     """
#     Test the speed of writing a large dataset
#     """
#     full_path = os.path.join(test_data_path, 'test_write_speed')
#     dataset = NDTiffDataset(full_path, summary_metadata={}, name='prefix')
#
#     image_height = 2048
#     image_width = 2028
#     N_images = 1000
#     image = np.random.randint(0, 2**16, size=(image_height, image_width), dtype=np.uint16)
#     MB_per_image = image_height * image_width * 2 / 1024 / 1024
#
#     image_write_times = []
#     start_time = time.time()
#     for t in tqdm(range(N_images)):
#         image_time = time.time()
#         axes = {'time': t}
#         dataset.put_image(axes, image, {'time_metadata': t})
#         image_write_times.append(time.time() - image_time)
#     dataset.finished_writing()
#
#     elapsed_time = time.time() - start_time
#     print(f"Time to write {N_images} images: ", elapsed_time,
#           "\nspeed: ", N_images * MB_per_image / elapsed_time, "MB/s")
#
#     return image_write_times, MB_per_image
#
#
# full_path = os.path.join('.', 'top_level')
#
# os.makedirs(full_path, exist_ok=True)
# for f in os.listdir(full_path):
#     os.remove(os.path.join(full_path, f))
#
# image_write_times, MB_per_image = writing_speed_test(full_path)
#
# MB_per_second_times = [MB_per_image / t for t in image_write_times]
#
# import matplotlib.pyplot as plt
# fig, ax = plt.subplots()
# ax.semilogy(MB_per_second_times)
# ax.set_xlabel('Image number')
# ax.set_ylabel('Speed (MB/s)')
# plt.show()
#
# shutil.rmtree(full_path)
